/*
 * sjabloon.c
 */

#include "sjabloon.h"

/******************************************************************************
 * sja_replace
 * Adds <content> IN FRONT OF the marker <node>.
 * This is intended, since there could be situations, where
 * multiple subsequent invokations of sja_replace on one node will be done.
 * Each individual marker will be processed once with the same action.
 *
 * returns:
 *	 0 	if there is no such token
 *	>0 	on success (number of tokens processed)
 *	-1 	on system error, i.e. realloc (see errno)	
 *****************************************************************************/

int sja_replace(sja_t *sja, const char const *node, const char const *content)
{
int pos=0;
int hits=0;

const size_t content_len=strlen(content);

while((pos=sja_tokat(sja,node,pos)))
	{
	/* 
	 * reserve space if needed
	 */
	if(sja->space <= (sja->len+content_len) )
		{
		unsigned int add=SJA_CHUNKED_SIZE(content_len);

		debug_int("old size",sja->space);
		debug_int("additional size",add);
					
		sja->buf=realloc(sja->buf,add+sja->space);

		if(!sja->buf) return -1;
		sja->space+=add;

		memset((sja->buf+sja->len+1),0,add);
		}

	pos--;

	debug_ptr("move from",sja->buf+pos);
	debug_ptr("move to",sja->buf+pos+content_len);
	debug_int("strlen(node)",strlen(node));
	debug_int("move region len",sja->len-pos);

	memmove
		(
		sja->buf+pos+content_len,
		sja->buf+pos,
		sja->len-pos
		);

	memcpy(sja->buf+pos,content,content_len);

	/*
	 * move pos to end of processed marker
	 * and update length of stored text
	 */
	pos=(pos+content_len+strlen(node)+2);
	sja->len+=content_len;

	debug_int("pos after move",pos);
	debug(sja->buf);

	hits++;
	}

return hits;
}

/******************************************************************************
 * sja_init_file
 * reads a file into a buffer and calls sja_init_string().
 * same return values as that function.
 *****************************************************************************/

sja_t *sja_init_file(const char const *fname)
{
char *buf=calloc(SJA_CHUNK_LEN,sizeof(char));
if(!buf) return NULL;

const int fd=open(fname, O_RDONLY);
if(fd<0) 
	{
	free(buf);
	return NULL;
	}

ssize_t wr=0;
ssize_t rd=SJA_CHUNK_LEN;	/* condition for the first run */
ssize_t buflen=SJA_CHUNK_LEN;

while(rd>=SJA_CHUNK_LEN)
	{
	/*
	 * realloc if needed
	 */
	if( (wr+SJA_CHUNK_LEN) > buflen)
		{
		debug_int("realloc",buflen+SJA_CHUNK_LEN);
		buf=realloc(buf,buflen+SJA_CHUNK_LEN);
		if(!buf)
			{
			debug("realloc() failed");
			close(fd);
			return NULL;
			}
		buflen+=SJA_CHUNK_LEN;
		}	

	rd=read(fd,buf+wr,SJA_CHUNK_LEN);

	if(rd<0)
		{
		debug("read() returned -1");
		free(buf);
		close(fd);
		return NULL;
		}

	wr+=rd;
	}

if(close(fd)<0)
	{
	debug("close() failed");
	free(buf);
	return NULL;
	}

debug("file reading completed");
debug_int("write pos ",wr);
debug_int("buf len ",buflen);

/*
 * alloc space if NULL termination won't fit
 */
if(wr+1 > buflen)
	{
	debug_int("realloc for \\0",SJA_CHUNKED_SIZE(buflen+1));
	buf=realloc(buf,SJA_CHUNKED_SIZE(buflen+1));
	if(!buf)
		{
		debug("unable to realloc for NULL");
		return NULL;
		}
	}
*(buf+wr)=0;

debug(buf);

sja_t *sja=sja_init_string(buf);
free(buf);

return sja;
}

/******************************************************************************
 * sja_init_string
 * Creates a sjabloon from a given NULL-terminated string.
 * Returns a pointer to a sjabloon or NULL on error (see errno).
 * The sjabloon has to be free'd by sja_destroy after use.
 *****************************************************************************/

sja_t *sja_init_string(const char const *data)
{
sja_t *sja=calloc(1,sizeof(sja_t));
if(!sja) return NULL;

const unsigned int len=SJA_CHUNKED_SIZE(strlen(data)+1);

sja->buf=calloc(len,sizeof(char));
if(!sja->buf)
	{
	free(sja);
	return NULL;
	}

strcpy(sja->buf,data);
sja->len=strlen(data);
sja->space=len;

return sja;
}


/******************************************************************************
 * sja_destroy
 * Frees all buffers from the sjabloon and returns 1.
 *****************************************************************************/

int sja_destroy(sja_t *sja)
{
if(sja->buf) free(sja->buf);
free(sja);
return 1;
}


/******************************************************************************
 * sja_tokat
 * Returns the position of the next char after '{' token 
 * in sja->buf or 0 if there is no such token, errno will be set to 0.
 *
 * If offset >0, search will begin at offset (to search sja->buf
 * in sequence). If offset is beyond the string saved in sjabloon, 
 * 0 will be returned and errno will be set to EFAULT.
 *
 * NOTE: 0 is a valid error return value, since no token(-name) can
 * begin at offset 0 without a preceeding {
 *****************************************************************************/

static unsigned int sja_tokat(const sja_t *sja, const char const *token, const unsigned int offset)
{
const char *tokptr=token;

unsigned int pos=0;
unsigned int cmpok=0;

if(offset>0) 
	{
	debug_int("offset defined as",offset);
	if(offset>=sja->len)
		{
		debug("... offset invalid");
		errno=EFAULT;
		return 0;
		}
	pos=offset;
	}

while(*(sja->buf+pos))
	{
	/*
	 * advance to next position of a token
	 */
	while(*(sja->buf+pos))
		{
		if(*(sja->buf+pos++)=='{') break;
		/* TODO: unroll */
		}

	if(*(sja->buf+pos)!=0)
		{
		debug("found {");

		tokptr=token;
		cmpok=0;

		/*
		 * compare next chars with token
		 * FIXME: reduce comparison
		 */
		while( (*tokptr && *(sja->buf+pos)) && (*tokptr==*(sja->buf+pos)) )
			{
			cmpok++;
			tokptr++;
			pos++;
			}

		debug("cmp done");

		if( (strlen(token)==cmpok) && *(sja->buf+pos)=='}')
			{
			return pos-cmpok;
			}
		pos++;
		}
	}

debug("token not found");
errno=0;
return 0;
}



/******************************************************************************
 *
 *****************************************************************************/

